/*** CPART.C - A utility to copy parts of a file or split and merge files.
 *   Version History
 *      1.02.19980822 - Initial public release.
 *      1.03.20151113 - Updates for 64-bit version (long -> long long);
 *                      Added "g" size.
 *
 *   (C) Fedon Kadifeli, 1998-2015.
 *
 *   Usage: See "printusage" function.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <conio.h>

enum bool { FALSE, TRUE } iseof;
char progname [FILENAME_MAX];


void prompt (const char * fmt, ...)
{
  va_list args;

  if (fmt != NULL) {
    va_start (args, fmt);
    vfprintf (stderr, fmt, args);
    va_end (args);
  }
  fprintf (stderr, "Press ESC to abort; any other key to continue... ");
  if (_getch () == 0x1B)
    exit (2);
  fprintf (stderr, "\r\t\t\t\t\t\t\t\t\t\r");
} /* prompt */


void printusage (const char * fmt, ...)
{

/* This function displays an error message before usage information */
/* and aborts. Note that, it does NOT return to the caller.         */

  va_list args;

  if (fmt != NULL) {
    va_start (args, fmt);
    vfprintf (stderr, fmt, args);
    va_end (args);
    prompt ("\n\nMore help follows... ");
  }
  fprintf (stderr,
"\n"
"A utility to copy parts of a file or split and merge files.\n"
"Version 1.03 (C) Fedon Kadifeli 1998-2015.\n\n"

"General format: %s -opt masfile partfile count startpos\n\n"

"Cases depending on <opt> are as follows:\n\n"

"%s -c[w] masfile partfile bytecount [startpos]\n"
"\tCopies <bytecount> bytes from <masfile> to <partfile> starting from\n"
"\t<startpos> (def. 0). If 'w' is appended to the 'c' option, the user\n"
"\tis prompted to press a key before each copy.\n\n"

"%s -s[w] masfile [partfile] [bytecount] [startpos]\n"
"\tSplits <masfile> into <partfile>s starting from <startpos> (def. 0).\n"
"\t.XXX is appended to each <partfile> name (<XXX> is a sequence number).\n"
"\tThe default for <partfile> is the name part of <masfile>. Each part\n"
"\tis <bytecount> bytes (def. 1d). If 'w' is appended to the 's' option,\n"
"\tthe user is prompted to press a key before each copy.\n\n",
    progname, progname, progname);
  prompt (NULL);
  fprintf (stderr,
"%s -m[w] masfile [partfile] [filecount]\n"
"\tMerges <filecount> <partfile>s into <masfile>. .XXX is appended to\n"
"\teach <partfile> name (<XXX> is a sequence number) before opening it.\n"
"\tThe default for <partfile> is the name part of <masfile>. The default\n"
"\tfor <filecount> is 999. If 'w' is appended to the 'm' option, the\n"
"\tuser is prompted to press a key before each copy.\n\n"

"Note: If no option is given, the command's first character is used by\n"
"default. For example, '-c' is the default option if the command is\n"
"called 'cpart'. If you rename cpart.exe as split.exe, then '-s' will\n"
"be assumed if <opt> is not given in the command line, etc.\n\n"

"<bytecount>, <startpos>, and <filecount> are decimal integer numbers\n"
"optionally followed by one of these suffices:\n"
"\tk for x Kilobyte (e.g., 800k = 819200)\n"
"\td for x 1420 x Kilobyte (e.g., 2d = 2908160)\n"
"\tm for x Megabyte (e.g., 10m = 10485760)\n"
"\tg for x Gigabyte (e.g., 5g = 5368709120)\n",
    progname);
  exit (1);
} /* printusage */


int fileexists (char *fn)
{
  FILE *fp = fopen (fn, "rb");

  if (fp != NULL) {
    fclose (fp);
    return TRUE;
  } else
    return FALSE;
} /* fileexists */


void checkfile (char *fn)
{
  int ch;

  if (fileexists(fn)) {
    fprintf (stderr, "File %s exists. Overwrite (y/n)? ", fn);
    do {
      ch = tolower (_getch ());
      if (ch == 'n') {
	fprintf (stderr, "n\nAborting...\n");
	exit (4);
      }
    } while (ch != 'y');
    fprintf (stderr, "y\n");
  }
} /* checkfile */


void file_error (char *fname)
{
  perror (fname);
  exit (3);
} /* file_error*/


FILE * fopen_ext (char *fname, char *fmode)
{
  FILE *fp;
 
  fp = fopen (fname, fmode);
  if (fp == NULL)
    file_error (fname);
  if (setvbuf (fp, NULL, _IOFBF, BUFSIZ*32) != 0)
    fprintf (stderr, "Warning: Cannot buffer %s!\n", fname);
/** Still, diskette operations on NT can be very slow **/
  return fp;
} /* fopen_ext */


char * getbasename (char *fn)
{
  static char pn[FILENAME_MAX];
  char *cp;

  cp = strrchr (fn, '\\');
  if (cp == NULL)
    cp = strrchr (fn, ':');
  strncpy (pn, (cp == NULL) ? fn : cp+1, sizeof(pn));
  cp = strrchr (pn, '.');
  if (cp != NULL)
    *cp = '\0';
  return pn;
/** Beware: there is only one buffer! **/
} /* getbasename */


long long cpart (char *fin, char *fon, long long count, long long startpos)
{
  FILE *fi;
  FILE *fo;
  long long wcount = 0;
  int c;

  fi = fopen_ext (fin, "rb");
  checkfile (fon);
  fo = fopen_ext (fon, "wb");
  fprintf (stderr, "Copying %s to %s (%lld bytes, starting from byte %lld)...\n",
    fin, fon, count, startpos);

  iseof = FALSE;

  while (startpos--) {
    c = getc (fi);
    if (c==EOF)
      break;
  } /* skip loop */

  while (count--) {
    c = getc (fi);
    iseof = (c==EOF);
    if (iseof)
      break;
    if (putc (c,fo) != c)
      file_error (fon);
    wcount++;
  } /* copy loop */

  if (!iseof)   /* look ahead */
    iseof = (getc(fi) == EOF);
  if (fclose (fi) != 0)
    file_error (fin);
  if (fclose (fo) != 0)
    file_error (fon);
  fprintf (stderr, "Wrote %lld bytes.\n", wcount);

  return wcount;
} /* cpart */


long long readnumpar (char *par)
{
  long long val = -1;
  char sfix = '\0';

  if (sscanf (par, "%lld%c", &val, &sfix) > 1)
    switch (tolower(sfix)) {
      case 'k' :
	val *= 1024L;
	break;
      case 'm' :
	val *= 1024L*1024L;
	break;
      case 'g' :
	val *= 1024L*1024L*1024L;
	break;
      case 'd' :
	val *= 1024L*1420L;
	break;
      default : 
	printusage ("Unknown number suffix: '%c'.", sfix);
    }
  if (val < 0)
    printusage ("'%s' is not recognized as a valid number.", par);
  return val;
} /* readnumpar */


int main (int argc, char **argv)
{
  int funchar = '\0';
  int waitchar = '\0';
  int progchar;

  strcpy (progname, getbasename(argv[0]));

  if (argc == 1)	/* no parameter */
    printusage (NULL);

  if ((argc>1) && (argv[1][0]=='-')) {    /* get option */
    int i, opt;
    for (i=1; (i<=2) && (argv[1][i]); i++)
      switch (opt = tolower (argv[1][i])) {
        case 'c' :
        case 's' :
        case 'm' :
          if (funchar)
	    printusage ("Only one of the 'c', 's', and 'm' options is allowed.");
          funchar = opt;
          break;
        case 'w' :
          if (waitchar)
	    printusage ("Only one occurrence of the 'w' option is allowed.");
          waitchar = opt;
          break;
	default:
          printusage ("Unknown option: '%c'", opt);
      } /* switch */
    if ((opt = tolower (argv[1][i])) != '\0')
      printusage ("Too many option characters starting from: '%c'.", opt);
    ++argv;
    --argc;
  } /* if */

  if (funchar == '\0')
    switch (progchar = tolower(progname[0])) {
      case 'c' :
      case 's' :
      case 'm' :
        funchar = progchar;
        break;
      default :
        printusage ("Cannot determine option. Specify one of the 'c', 's', 'm' options.");
    } /* switch */

  switch (funchar) {

    case 'c' :         /* copy part of file */

      if ((argc==4) || (argc==5)) {

	char *fmas, *fpart;
	long long count = 0;
	long long startpos = 0;

	fmas = *++argv;
	fpart = *++argv;
	++argv;
	count = readnumpar (*argv);
	++argv;
	if (*argv != NULL)
	  startpos = readnumpar (*argv);
        if (waitchar)
	  prompt ("About to copy %s... ", fpart);
	cpart (fmas, fpart, count, startpos);

      } else

	printusage ("Incorrect number of parameters for copy part operation.");

      break;

    case 's' :        /* split file */

      if ((argc>=2) && (argc<=5)) {

	char *fmas, *fpart, fparte[FILENAME_MAX+5];
	long long count, startpos, cc;
	int i;

	++argv;
	fmas = *argv++;
	if (*argv != NULL)
	  fpart = *argv++;
	else
	  fpart = getbasename (fmas);
	if (*argv != NULL)
	  count = readnumpar (*argv++);
	else
	  count = readnumpar ("1d");
	if (*argv != NULL)
	  startpos = readnumpar (*argv);
	else
	  startpos = 0;
	for (i=1; i<1000; i++) {
	  sprintf (fparte, "%s.%03d", fpart, i);
	  if (waitchar)
	    prompt ("About to copy %s... ", fparte);
	  cc = cpart (fmas, fparte, count, startpos);
	  if (iseof || (cc <= 0))
	    break;
	  startpos += cc;
	}
	if (i==1000)
	  fprintf (stderr, "Too many files to be copied. Aborting...\n");

      } else

	printusage ("Incorrect number of parameters for split file operation.");

      break;

    case 'm' :        /* merge file */

      if ((argc>=2) && (argc<=4)) {

	char *fmas, *fpart, fparte[FILENAME_MAX+5];
	FILE *fm, *fp;
	int count, i, c;
	long long cc;

        ++argv;
	fmas = *argv++;
        if (*argv != NULL)
	  fpart = *argv++;
	else
	  fpart = getbasename (fmas);
        if (*argv != NULL) {
          count = readnumpar (*argv);
          if (count>=1000)
            printusage ("File count must be smaller than 1000.");
	  else if (count<=0)
            printusage ("File count must be greater than 0.");
        } else
          count = 1000;
        checkfile (fmas);
	fm = fopen_ext (fmas, "wb");
	for (i=1; i<=count; i++) {
	  sprintf (fparte, "%s.%03d", fpart, i);
	  if (waitchar)
	    prompt ("About to copy %s... ", fparte);
          if ((count == 1000) && (i>1) && !fileexists (fparte))
            break;
	  fp = fopen_ext (fparte, "rb");
	  fprintf (stderr, "Appending %s to %s...\n", fparte, fmas);
	  cc = 0;
	  while ((c=getc(fp))!=EOF) {
	    cc++;
	    if (putc (c,fm) != c)
              file_error (fmas);
	  }
	  fprintf (stderr, "Appended %lld bytes.\n", cc);
	  if (fclose (fp) != 0)
            file_error (fparte);
	}
	if (fclose (fm) != 0)
          file_error (fmas);

      } else

	printusage ("Incorrect number of parameters for merge file operation.");

      break;
      
  } /* switch */

  return 0;
} /* main */
